/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.IOException;
import java.io.Writer;

import com.ibm.jzos.ByteUtil;
import com.ibm.jzos.fields.AssemblerDatatypeFactory;
import com.ibm.jzos.fields.BinaryAsIntField;
import com.ibm.jzos.fields.BinaryAsLongField;
import com.ibm.jzos.fields.StringField;

/**
 * This class maps the ADATA DCDS Operand block
 * <p>
 * @see ADataDCDSRecord
 */
public class ADataDCDSOperand {
	
	public static final int BUFFER_LEN = 0x001F; 
	
	private byte[] bytes;
	private int bufOffset;

	protected static AssemblerDatatypeFactory factory = new AssemblerDatatypeFactory();

	// *  The following 9 fields make up a DCDS Operand Block, and this   
	// *  block is repeated, not contiguously, for a count supplied in    
	// *  ADDCDS_OPS_NUM.                                                 
	// *                                                                  
	// *  The offset of the first block, from the start of the ADATA      
	// *  record, is supplied in ADDCDS_OPS_OFF.  Each block contains     
	// *  the offset to the next block, from the start of the ADATA       
	// *  record.  The blocks end when the next offset is zero.           
	// *                                                                  
	// ADDCDS_OPERAND      DSECT                                          
	// ADDCDS_NEXT_OP_OFF  DS    FL4        Offset of next operand block 
	private static BinaryAsIntField _ADDCDS_NEXT_OP_OFF = (BinaryAsIntField)factory.getBinaryField(4, true);

	// ADDCDS_LOCTR        DS    FL4        Location counter              
	private static BinaryAsIntField _ADDCDS_LOCTR = (BinaryAsIntField)factory.getBinaryField(4, true);

	// ADDCDS_DUP          DS    FL4        The duplication factor        
	private static BinaryAsIntField _ADDCDS_DUP = (BinaryAsIntField)factory.getBinaryField(4, true);
	
	// ADDCDS_BIT_OFFSET   DS    XL1        Offset within byte for B const
	private static BinaryAsIntField _ADDCDS_BIT_OFFSET = (BinaryAsIntField)factory.getBinaryField(1, false);
	
	// ADDCDS_TYPE_ATTR    DS    XL1        Type attribute                
	private static StringField _ADDCDS_TYPE_ATTR = factory.getStringField(1, false);

	// ADDCDS_TYPE_EXT     DS    CL1        Type extension
	private static StringField _ADDCDS_TYPE_EXT = factory.getStringField(1, false);
	
	// ADDCDS_PROGRAM_TYPE DS    XL4        Program Type                  
	private static BinaryAsLongField _ADDCDS_PROGRAM_TYPE = (BinaryAsLongField)factory.getBinaryField(4, false);

	//                     DS    XL4        Reserved                      
	static { factory.getBinaryField(4, false); }
	
	// ADDCDS_VALUES_NUM   DS    FL4        Number of nominal values      
	private static BinaryAsIntField _ADDCDS_VALUES_NUM = (BinaryAsIntField)factory.getBinaryField(4, true);

	// ADDCDS_VALUES_OFF   DS    FL4        Offset of first nominal value 
	private static BinaryAsIntField _ADDCDS_VALUES_OFF = (BinaryAsIntField)factory.getBinaryField(4, true);

	// ADDCDS_OPERAND_EQU_LEN    EQU  *-ADDCDS_OPERAND                    

	/**
	 * Construct an instance on an existing byte array.
	 * @param buffer the byte array used as backing memory for the structure
	 */
	public ADataDCDSOperand(byte[] buffer, int bufOffset) {
		if (buffer.length < BUFFER_LEN) {
			throw new IllegalArgumentException("buffer length");
		}
		bytes = buffer;
		this.bufOffset = bufOffset;
	}

	public int getNEXT_OP_OFF() {
		return _ADDCDS_NEXT_OP_OFF.getInt(bytes, bufOffset);
	}
	
	public int getLOCTR() {
		return _ADDCDS_LOCTR.getInt(bytes, bufOffset);
	}
	
	public long getPROGRAM_TYPE() {
		return _ADDCDS_PROGRAM_TYPE.getLong(bytes, bufOffset);
	}
	
	public int getDUP() {
		return _ADDCDS_DUP.getInt(bytes, bufOffset);
	}
	
	public int getBIT_OFFSET() {
		return _ADDCDS_BIT_OFFSET.getInt(bytes, bufOffset);
	}
	
	public String getTYPE_ATTR() {
		return _ADDCDS_TYPE_ATTR.getString(bytes, bufOffset);
	}
	
	public String getTYPE_EXT() {
		return _ADDCDS_TYPE_EXT.getString(bytes, bufOffset);
	}
	
	public int getVALUES_NUM() {
		return _ADDCDS_VALUES_NUM.getInt(bytes, bufOffset);
	}

	public int getVALUES_OFF() {
		return _ADDCDS_VALUES_OFF.getInt(bytes, bufOffset);
	}

	public ADataDCDSOperand getNextOperand() {
		int off = getNEXT_OP_OFF();
		return new ADataDCDSOperand(bytes, off);
	}
	
	public ADataDCDSValue getFirstValue() {
		if (getVALUES_NUM() <= 0) {
			return null;
		}
		int off = getVALUES_OFF();
		return new ADataDCDSValue(bytes, off);
	}
	
	/**
	 * Write a (debug) dump of the receiver on a Writer.
	 * @param writer
	 * @throws IOException
	 */
	public void dumpOn(Writer writer) throws IOException {
		writer.write("  Operand: LOCTR=0x");
		writer.write(ByteUtil.toHexString(getLOCTR(), 4));
		writer.write("  DUP=");
		writer.write(Integer.toString(getDUP()));
		writer.write("  TYPE=");
		writer.write(getTYPE_ATTR());
		writer.write("  EXT=");
		writer.write(getTYPE_EXT());
		writer.write("\n");
	}
}
