/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.IOException;
import java.io.Writer;

import com.ibm.jzos.ByteUtil;
import com.ibm.jzos.fields.AssemblerDatatypeFactory;
import com.ibm.jzos.fields.BinaryAsIntField;

/**
 * Maps a AData DCDS record VALUE block
 * <p>
 * @see ADataDCDSRecord
 */
public class ADataDCDSValue {
	
	public static final int BUFFER_LEN = 0x0010; 
	
	private byte[] bytes;
	private int bufOffset;

	protected static AssemblerDatatypeFactory factory = new AssemblerDatatypeFactory();

	// *  The following 4 fields make up a DCDS Value Block, and this     
	// *  block is repeated, not contiguously, for a count supplied in    
	// *  ADDCDS_VALUES_NUM.                                              
	// *                                                                  
	// *  The offset of the first block, from the start of the ADATA      
	// *  record, is supplied in ADDCDS_VALUES_OFF.  Each block contains  
	// *  the offset to the next block, from the start of the ADATA       
	// *  record.  The blocks end when the next offset is zero.           
	// *                                                                  
	// ADDCDS_VALUE        DSECT                                          
	// ADDCDS_NEXT_VAL_OFF DS    FL4        Offset of next value block   
	private static BinaryAsIntField _ADDCDS_NEXT_VAL_OFF = (BinaryAsIntField)factory.getBinaryField(4,true);
	
	// ADDCDS_OBJECT_OFF   DS    FL4        Offset of the nominal value   
	private static BinaryAsIntField _ADDCDS_OBJECT_OFF = (BinaryAsIntField)factory.getBinaryField(4,true);

	// ADDCDS_BYTE_LEN     DS    FL4        # of bytes in the nominal val 
	private static BinaryAsIntField _ADDCDS_BYTE_LEN = (BinaryAsIntField)factory.getBinaryField(4,true);

	// ADDCDS_BIT_LEN      DS    FL4        # of Bits if B type constant  
	private static BinaryAsIntField _ADDCDS_BIT_LEN = (BinaryAsIntField)factory.getBinaryField(4,true);
	
	// ADDCDS_VALUE_EQU_LEN      EQU  *-ADDCDS_VALUE
	// *                                                                  
	// *  The following field may be present for each occurance of a      
	// *  DCDS Value Block, and can be accessed using the offset and      
	// *  length provided.  The _OFF gives the offset of the related      
	// *  field, from the start of the ADATA record, and the _LEN is the  
	// *  length of the related field.  Only one of the lengths will be   
	// *  provided; either the byte length or the bit length.  An offset  
	// *  of zero indicates that the related field is not available.      
	// *                                                                  
	// *  The field will not be present for a DC with a duplication       
	// *  factor of zero, or a DS or a DXD.                               
	// *                                                                  
	// ADDCDS_OBJECT       DS    0X         Nominal value                 

	/**
	 * Construct an instance on an existing byte array.
	 * @param buffer the byte array used as backing memory for the structure
	 */
	public ADataDCDSValue(byte[] buffer, int bufOffset) {
		if (buffer.length < BUFFER_LEN) {
			throw new IllegalArgumentException("buffer length");
		}
		bytes = buffer;
		this.bufOffset = bufOffset;
	}
	
	public int getNEXT_VAL_OFF() {
		return _ADDCDS_NEXT_VAL_OFF.getInt(bytes, bufOffset);
	}
	
	public int getOBJECT_OFF() {
		return _ADDCDS_OBJECT_OFF.getInt(bytes, bufOffset);
	}
	
	public int getBYTE_LEN() {
		return _ADDCDS_BYTE_LEN.getInt(bytes, bufOffset);
	}
	
	public int getBIT_LEN() {
		return _ADDCDS_BIT_LEN.getInt(bytes, bufOffset);
	}

	/**
	 * Answer the next linked ADataDCDSValue
	 * @return ADataDCDSValue
	 */
	public ADataDCDSValue getNextValue() {
		int off = getNEXT_VAL_OFF();
		if (off == 0) return null;
		return new ADataDCDSValue(bytes, off);
	}

	/**
	 * Write a (debug) dump of the receiver
	 * @param writer
	 * @throws IOException
	 */
	public void dumpOn(Writer writer) throws IOException {
		writer.write("  Value: BYTE_LEN=");
		writer.write(Integer.toString(getBYTE_LEN()));
		writer.write("  BIT_LEN=");
		writer.write(Integer.toString(getBIT_LEN()));
		int olen = getNEXT_VAL_OFF() - getOBJECT_OFF();
		if (olen > 0) {
			writer.write("  Object: 0x");
			writer.write(ByteUtil.toHexString(bytes, bufOffset+getOBJECT_OFF(), olen )); 
		}
		writer.write("\n");
	}
	
}
