/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import com.ibm.jzos.fields.AssemblerDatatypeFactory;
import com.ibm.jzos.fields.BigDecimalAccessor;
import com.ibm.jzos.fields.BigIntegerAccessor;
import com.ibm.jzos.fields.ByteArrayField;
import com.ibm.jzos.fields.DatatypeFactory;
import com.ibm.jzos.fields.DoubleAccessor;
import com.ibm.jzos.fields.Field;
import com.ibm.jzos.fields.FloatAccessor;
import com.ibm.jzos.fields.IntAccessor;
import com.ibm.jzos.fields.LongAccessor;
import com.ibm.jzos.fields.StringField;
import com.ibm.jzos.recordgen.LibraryUtil;
import com.ibm.jzos.recordgen.LineFormatter;

/**
 * Generates the source code for a Java class that maps a byte array matching
 * the description given by an Assembler CSECT or DSECT.  The input to this
 * process is either an ADATA file created by the IBM High-level Assembler
 * or an XML file generated by {@link RecordSpecXMLGenerator}
 * <p>
 * See the sample JCL member "DSECTGEN" for example JCL for executing the 
 * assembler and converting the ADATA output to a Java source file. 
 * <p>
 * @see #main(String[]) main() method for information on running this class
 */
public class RecordClassGenerator {

	private static final String _TYPES_PACKAGE_NAME = "com.ibm.jzos.fields";
	private RecordSpec recordSpec;
	
	private LineFormatter formatter;
	private AssemblerDatatypeFactory factory = new AssemblerDatatypeFactory();	
	
	// control properties
	/** The package name of the output class */
	private String packageName = "mypackage";  

	/** The name of the output class.  If null defaults to the top-level record name.
	 * @see getClassName()
	 */
	private String className = null;

	/** Whether field setter methods are generated */
	private boolean genSetters = true;

	/** Whether instances variables are used as field cache values */
	private boolean genCache = true;

	/** Whether static fields are generated protected (true) or public(false) */
	private boolean genProtectedFields = true;

	/** Whether a class that uses bufOffset if generated */
	private boolean useBufOffset = false;

	/** The default setting for {@link DatatypeFactory#setStringTrimDefault(boolean)} */
	private boolean stringTrimDefault = false;

	/** Whether sourcelines are used as javadoc for getters and setters */
	private boolean genAccessorJavadoc = false;
	
	/** Whether getter and setter method names are generated as camelCase */
	private boolean camelCaseAccessors = true;
	
	/**
	 * Main method / driver for generating Java Classes from Assembler ADATA files
	 * or XML.
	 * <p>
	 * Arguments consist of the following optional "name=value" pairs:  
	 * <ul>
	 * <li>adatafile=//MVS.DSN  (default: //DD:SYSADATA)</li>
	 * <li>xmlfile=//MVS.DSN    (default: none)</li>
	 * <li>outputDir=dirname    (default: writes file to System.out (stdout))</li>
	 * <li>section=DSECTNAME    (default: process the first CSECT or DSECT found)</li>
	 * <li>package=packageName  (default: "mypackage")</li>
	 * <li>class=className      (default: name of the C/DSECT section)</li>
	 * <li>genSetters=true|false (default: true)</li>
	 * <li>genCache=true|false     {default: true)</li>
	 * <li>bufOffset=true|false (default: false)</li>
	 * <li>stringTrim=true|false (default: false)</li>
	 * <li>commentsPrecedeDeclaration=true|false (default: false)</li>
	 * <li>genAccessorJavadoc=true|false (default: false)</li>
	 * <li>camelCaseAccessors=true|false (default: true)</li>
	 * </ul>
	 * <p> 
	 * If <code>outputDir</code> is given, the output file is created in the package-qualified
	 * subdirectory with a file name "Classname.java".  For example, if 
	 * <code>outputDir=/u/myid</code> and <code>package=com.myco.cobrecs</code> and
	 * <code>class=MyRecord</code> or implicitly determined from the copybook record 
	 * <code>MY-RECORD</code>, then the output file created will be: 
	 * <code>/u/myid/com/myco/cobrecs/MyRecord.java</code>
	 * <p>
	 * If the given <code>adatafile</code> is a pathname not starting with "//", then it is assumed
	 * to be the name of a binary file in the USS (or workstation) filesystem containing the 
	 * binary SYSADATA file.  In order to process SYSADATA files on the workstation, 
	 * they must be downloaded with IBM "RDWs" included.  This can be accomplished using FTP 
	 * along with the "bin" and "quote site rdw" commands".
	 * <p>
	 * If the given <code>adatafile</code> is a PDS name without a member or a directory name,
	 * then all members of the PDS (or files in the directory) are processed.  For this to work
	 * properly, the <code>outputDir</code> option must be specified but the 
	 * <code>class</code> and <code>symbol</code> options must not be specified so that the copybook top-level
	 * record name is used to generate unique classnames for each class.   This feature allows
	 * for one invocation of the RecordClassGenerator to generate Java classes for an
	 * entire PDS library of SYSADATA files.
	 * <p> 
	 * @param args key=value strings as described above.
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		
		String adatafile = "//DD:SYSADATA";
		boolean adatafileGiven = false;
		String xmlfile = null;
		String outputdir = null;
		String section = null;
		String packageName = "mypackage";
		String className = "MyClass";
		boolean useBufOffset = false;
		boolean genSetters = true;
		boolean genCache = true;
		boolean stringTrim = false;
		boolean commentsPrecedeDeclaration = false;
		boolean genAccessorJavadoc = false;
		boolean camelCaseAccessors = true;
		
		for (int i = 0; i<args.length; i++) {
			StringTokenizer strtok = new StringTokenizer(args[i],"=");
			if (strtok.countTokens() != 2) {
				throw new IllegalArgumentException("Argument '" + args[i] + "' not in the form: name=value");
			}
			String name = strtok.nextToken();
			String value = strtok.nextToken();
			if ("adatafile".equalsIgnoreCase(name)) {
				adatafile = value;
				adatafileGiven = true;
			} else if ("xmlfile".equalsIgnoreCase(name)) {
					xmlfile = value;
			} else if ("outputdir".equalsIgnoreCase(name)) {
				outputdir = value;
			} else if ("section".equalsIgnoreCase(name)) {
				section = value;
			} else if ("package".equalsIgnoreCase(name)) {
				packageName = value;
			} else if ("class".equalsIgnoreCase(name)) {
				className = value;
			} else if ("genAccessorJavadoc".equalsIgnoreCase(name)) {
				genAccessorJavadoc = "true".equalsIgnoreCase(value);
			} else if ("genSetters".equalsIgnoreCase(name)) {
				genSetters = "true".equalsIgnoreCase(value);
			} else if ("genCache".equalsIgnoreCase(name)) {
				genCache = "true".equalsIgnoreCase(value);
			} else if ("bufOffset".equalsIgnoreCase(name)) {
				useBufOffset = "true".equalsIgnoreCase(value);
			} else if ("stringTrim".equalsIgnoreCase(name)) {
				stringTrim = "true".equalsIgnoreCase(value);
			} else if ("commentsPrecedeDeclaration".equalsIgnoreCase(name)) {
				commentsPrecedeDeclaration = "true".equalsIgnoreCase(value);
			} else if ("camelCaseAccessors".equalsIgnoreCase(name)) {
				camelCaseAccessors = "true".equalsIgnoreCase(value);
			} else {
				throw new IllegalArgumentException("Argument '" + name + "' not recognized.");
			}
		}
		
		if (xmlfile != null && adatafileGiven) {
			throw new IllegalArgumentException("adatafile= and xmlfile= keywords are mutually exclusive");
		}
		if (xmlfile != null && section != null ) {
			throw new IllegalArgumentException("section= keyword may not be specified with xmlfile= keyword");
		}
		if (xmlfile != null && commentsPrecedeDeclaration ) {
			throw new IllegalArgumentException("commentsPrecedeDeclaration= keyword may not be specified with xmlfile= keyword");
		}
		// Create a generator and configure it with the control arguments
		RecordClassGenerator rcg = new RecordClassGenerator();
		rcg.setClassName(className);
		rcg.setPackageName(packageName);
		rcg.setUseBufOffset(useBufOffset);
		rcg.setGenSetters(genSetters);
		rcg.setGenCache(genCache);
		rcg.setStringTrimDefault(stringTrim);
		rcg.setGenAccessorJavadoc(genAccessorJavadoc);
		rcg.setCamelCaseAccessors(camelCaseAccessors);
		
		if (xmlfile != null) {
			processXmlFile(xmlfile, outputdir, rcg);
		} else {
			processLibrary(adatafile, outputdir, rcg, section, commentsPrecedeDeclaration);
		}
	}
	
	/**
	 * Process all of the members of a given adatafile if it is a PDS, or all of the
	 * files if it is a directory, or a single file otherwise.
	 */
	private static void processLibrary(String adatafile, String outputDir, RecordClassGenerator rcg, String section, boolean commentsPrecedeDeclaration) throws IOException {

		// Get the the fully qualified member file names in the library, 
		// which would be a single name if the adatafile was not a library or directory
		List files = LibraryUtil.getLibraryMembers(adatafile);
		
		if (files.size() == 0) {
			throw new IllegalArgumentException("The given adatafile PDS or directory was empty");
		}
		if (files.size() > 1) {
			if (rcg.getClassName() != null) {
				System.err.println("class=className argument cannot be specified if a adatafile PDS or directory is processed and is ignored");
				rcg.setClassName(null);
			}
			if (section != null) {
				System.err.println("section=name argument cannot be specified if a adatafile PDS or directory is processed and is ignored");
				section = null;
			}
			if (outputDir == null) {
				throw new IllegalArgumentException("The outputDir= argument must be supplied when processing a PDS or directory");
			}
		}
		for (Iterator i=files.iterator(); i.hasNext(); ) {
			String filename = (String)i.next();
			processFile(filename, outputDir, rcg, section, commentsPrecedeDeclaration);
		}	
	}
	
	
	
	/**
	 * Process a single SYSADATA file
	 * @param adatafile the file name.  If starts with "//", then it is assumed to be 
	 * a z/OS dataset
	 * @param outputDir the output directory name where the output class is written, 
	 * under its package-qualified subdirectories
	 * @param rcg the RecordClassGenerator which has the class and package name
	 * @param sectionName the name of the CSECT or DSECT to process.  
	 * If null, then process the first encountered.
	 * @param commentsPrecedeDeclaration if true, assume comments precede symbol or
	 * DCDS declaration in the DSECT.  If false, assume comments follow the declaration.
	 */
	private static void processFile(
			String adatafile, 
			String outputDir, 
			RecordClassGenerator rcg, 
			String sectionName, 
			boolean commentsPrecedeDeclaration) {
		
		BufferedWriter wtr = null;
		try {
			// First, a RecordSpecBuilder converts the selected symbol of the ADATA
			// file into a RecordSpec model.
			adatafile = LibraryUtil.resolveLeadingTilde(adatafile);
			RecordSpecBuilder specBuilder = new RecordSpecBuilder();
			specBuilder.setSelectedSectionName(sectionName);
			specBuilder.setCommentsPrecedeDeclaration(commentsPrecedeDeclaration);
			RecordSpec recordSpec = specBuilder.process(adatafile);
			rcg.setRecordSpec(recordSpec);
		
			// Generate Java source to System.out
			File f = openOutputFile(rcg, outputDir);
			String outName;
			if (f != null) {
				wtr = new BufferedWriter(new FileWriter(f));
				outName = f.getCanonicalPath();
			} else {
				wtr = new BufferedWriter(new OutputStreamWriter(System.out));
				outName = "System.out (stdout)";
			}
			System.out.println("Generating " + adatafile + " to: " + outName);
			rcg.generateJavaSourceOn(wtr);
			
		} catch (Throwable t) {
			System.err.println("Error while processing file: " + adatafile + "\n\t" + t);
			t.printStackTrace();
		} finally {
			try { // to close the writer
				if (wtr != null) wtr.close(); 
			} catch(Throwable ignore) {}
		}
	}
	
	/**
	 * Process a single XML input file
	 * @param xmlfile the file name.  If starts with "//", then it is assumed to be 
	 * a z/OS dataset
	 * @param outputDir the output directory name where the output class is written, 
	 * under its package-qualified subdirectories
	 * @param rcg the RecordClassGenerator which has the class and package name
	 */
	private static void processXmlFile(
			String xmlfile, 
			String outputDir, 
			RecordClassGenerator rcg ) {
		
		BufferedWriter wtr = null;
		try {
			// First, a RecordSpecBuilder converts the selected symbol of the ADATA
			// file into a RecordSpec model.
			xmlfile = LibraryUtil.resolveLeadingTilde(xmlfile);
			RecordSpecXMLParser specParser = new RecordSpecXMLParser();
			RecordSpec recordSpec = specParser.process(xmlfile);
			rcg.setRecordSpec(recordSpec);
		
			// Generate Java source to System.out
			File f = openOutputFile(rcg, outputDir);
			String outName;
			if (f != null) {
				wtr = new BufferedWriter(new FileWriter(f));
				outName = f.getCanonicalPath();
			} else {
				wtr = new BufferedWriter(new OutputStreamWriter(System.out));
				outName = "System.out (stdout)";
			}
			System.out.println("Generating " + xmlfile + " to: " + outName);
			rcg.generateJavaSourceOn(wtr);
			
		} catch (Throwable t) {
			System.err.println("Error while processing file: " + xmlfile + "\n\t" + t);
			t.printStackTrace();
		} finally {
			try { // to close the writer
				if (wtr != null) wtr.close(); 
			} catch(Throwable ignore) {}
		}
	}	
	/**
	 * Return a File used to write to a generated output java source
	 * file, given the outputDir and class and package names from the given RecordClassGenerator.
	 * If the directory name is null, then return a BufferedWriter to System.out.
	 * @param rcg RecordClassGenerator used to determine the className and packageName; 
	 * @param outputDir the name of the output directory
	 */
	public static File openOutputFile(RecordClassGenerator rcg, String outputDir) throws IOException {
		
		if (outputDir == null) {
			return null; 
		}
		if (rcg.getClassName() == null) {
			throw new IllegalArgumentException("Given RecordClassGenerator does not have a className or sectionName");
		}
		outputDir = LibraryUtil.resolveLeadingTilde(outputDir);
		File dir = new File(outputDir);
		if (!dir.exists()) {
			throw new IllegalArgumentException("Output directory \"" + outputDir + "\" does not exist");
		}
		if (!dir.isDirectory()) {
			throw new IllegalArgumentException("Output directory \"" + outputDir + "\" is not a directory");
		}
		String packageName = rcg.getPackageName().replace('.', '/');
		dir = new File(dir, packageName);
		if (!dir.exists()) {
			if (!dir.mkdirs()) {
				throw new IOException("Unable to make directory: " + dir.getCanonicalPath());
			}
		}
		return  new File(dir, rcg.getClassName() + ".java");
		
	}
	
	/**
	 * Construct an empty instance.
	 * 
	 * @see #setRecordSpec(RecordSpec)
	 */
	public RecordClassGenerator() {
	}
	
	/**
	 * Construct an instance with a given RecordSpec
	 * @param recordSpec the RecordSpec
	 */
	public RecordClassGenerator(RecordSpec recordSpec) {
		this.recordSpec = recordSpec;
	}
	
	/**
	 * Sets the RecordSpec to process
	 */
	public void setRecordSpec(RecordSpec recordSpec) {
		this.recordSpec = recordSpec;
	}
	
	/**
	 * Generate Java source to the given Writer.
	 * @param writer
	 * @throws IOException
	 */
	public void generateJavaSourceOn(Writer writer) throws IOException {
		
		this.formatter = new LineFormatter(writer);
		genHeader();
		genStaticFieldLayout();
		genInstanceVariables();
		genConstructors();
		genBufferAccessors();
		genSettersAndGetters();
		genTrailer();
	}
	
	/**
	 * Answer the unqualified Java class name that is generated.
	 * If this is not given explicitly, then the selected C/DSECT 
	 * symbol name is used to generate a Java class name. 
=	 * 
	 * @return String the unqualified ClassName
	 */
	public String getClassName() {
		if (className != null) {
			return className;
		} else if (recordSpec == null) {
			return null;
		} else {
			String sectionName = recordSpec.getSectionName();
			return getInstanceJavaName(sectionName, 0, true);
		}
	}
	
	/**
	 * @see #getClass()
	 */
	public void setClassName(String className) {
		this.className = className;
	}

	/**
	 * Answer the Java package name for the generated Class.
	 * @return String
	 */
	public String getPackageName() {
		return packageName;
	}	

	/**
	 * @see #getPackageName()
	 */
	public void setPackageName(String packageName) {
		this.packageName = packageName;
	}

	/**
	 * Return whether or not javadoc comments are to be generated from the
	 * underlying sourcelines.
	 */
	public boolean isGenAccessorJavadoc() {
		return genAccessorJavadoc;
	}
	
	/**
	 * @see #isGenAccessorJavadoc()
	 */
	public void setGenAccessorJavadoc(boolean generateAccessorJavadoc) {
		this.genAccessorJavadoc = generateAccessorJavadoc;
	}

	/**
	 * Answer whether code is generated to use instance variables to
	 * cache field values.
	 */
	public boolean isGenCache() {
		return genCache;
	}
	
	/**
	 * @see #isGenCache()
	 */
	public void setGenCache(boolean value) {
		genCache = value;
	}

	/**
	 * Answer whether static Fields are generated "protected" (true)
	 * or "public" (false).
	 */
	public boolean isGenProtectedFields() {
		return genProtectedFields;
	}
	
	/**
	 * @see #isGenProtectedFields()
	 */
	public void setGenProtectedFields(boolean value) {
		genProtectedFields = value;
	}
	
	/**
	 * Answer whether Java "set" methods should be generated for generated fields.
	 * @return boolean
	 */
	public boolean isGenSetters() {
		return genSetters;
	}

	/**
	 * @see #isGenSetters()
	 */
	public void setGenSetters(boolean genSetters) {
		this.genSetters = genSetters;
	}

	/**
	 * Answer whether generated Java source code will include buffer offsets
	 * into the underlying byte array.
	 * @return boolean
	 */
	public boolean isUseBufOffset() {
		return useBufOffset;
	}

	/**
	 * @see #isUseBufOffset()
	 */
	public void setUseBufOffset(boolean useBufOffset) {
		this.useBufOffset = useBufOffset;
	}
	
	/**
	 * Answer the default value for StringField trimming.
	 */
	public boolean getStringTrimDefault() {
		return stringTrimDefault;
	}
	
	/**
	 * Set the default trim setting for StringFields
	 * @see #getStringTrimDefault()
	 */
	public void setStringTrimDefault(boolean value) {
		this.stringTrimDefault = value;
	}

	/**
	 * Answer whether generated Java source code getter and setter methods
	 * will have camelCase names (the default) or take the casing of the
	 * ADATA label.
	 * @return boolean
	 */
	public boolean isCamelCaseAccessors() {
		return camelCaseAccessors;
	}

	/**
	 * Set the behavior for getter and setter method generation.
	 * @see #isCamelCaseAccessors()
	 */
	public void setCamelCaseAccessors(boolean camelCaseAccessors) {
		this.camelCaseAccessors = camelCaseAccessors;
	}

	/**
	 * Generate a Java static final constant, which may be an absolute Integer, String, 
	 * or relative offset the the current factory's offset.
	 * @param constantSpec the ConstantSpec
	 * @throws IOException
	 */
	protected void genConstant(ConstantSpec constantSpec) throws IOException {
		genSourceLineComments(constantSpec);
		
		Object value = constantSpec.getValue();
		String primType;
		String valDecl;
		int intVal = 0;
		char charVal = 0;
		String javaName = getStaticJavaName(constantSpec.getLabel(), 0);
		
		if (value instanceof Integer) {
			intVal = ((Integer)value).intValue();
			primType = "int";
			valDecl = "0x" + Integer.toHexString(intVal);
		} else if (value instanceof Character) {
			charVal = ((Character)value).charValue();
			primType = "char";
			valDecl = "'" + Character.toString(charVal) + "'";
		} else {
			return;  // bail if we don't support
		}
		formatter.print("public static final % % = ", primType, javaName);
		
		if (constantSpec.isRelocatable()) {
			int bias = intVal - factory.getOffset();
			formatter.print("factory.getOffset()");
			if (bias != 0) {
				formatter.print(" + 0x%", Integer.toHexString(bias));
			}
		} else {
			formatter.print(valDecl);
		}
		formatter.println(";");
	}
	
	/**
	 * Generate constructors
	 * @throws IOException
	 */
	protected void genConstructors() throws IOException {
		
		formatter.println();
		formatter.println("public %() {", getClassName());
		formatter.println("this._byteBuffer = new byte[_DEFAULT_LEN];");
		formatter.println("}");
		
		formatter.println();
		formatter.println("public %(byte[] buffer%) {",
							getClassName(),
							useBufOffset ? ", int bufOffset" : "");
		
		formatter.println("this._byteBuffer = buffer;");
		if (useBufOffset) {
			formatter.println("this._byteBufferOffset = bufOffset;");
		}
		formatter.println("}");
	}
	
	/**
	 * Generate a file header for the class
	 * @throws IOException
	 */
	protected void genHeader() throws IOException {
		formatter.println("package %;", getPackageName());
		formatter.println("import %.*;", _TYPES_PACKAGE_NAME);
		if (recordSpec.containsBigDecimalsOrIntegers()) {
			formatter.println("import java.math.*;");
		}
		formatter.println();
		
		formatter.println("// Generated by % on: %",
							getClass().getName(),
							Calendar.getInstance().getTime());
		formatter.println();
		
		formatter.println("public class % {",
							getClassName());
		
	}
	
	/**
	 * Generate an instance variable for a given FieldSpec
	 * @param fieldSpec the FieldSpec
	 * @throws IOException
	 */
	protected void genInstanceVariable(FieldSpec fieldSpec) throws IOException {

		Field field = fieldSpec.getField();
		String label = fieldSpec.getLabel();
		if (label == null) return; 
		
		String javaName = getInstanceJavaName(label, fieldSpec.getOperandIndex(), false);
		String fieldName = getStaticJavaName(label, fieldSpec.getOperandIndex());
		
		String ivClassName = getIvClassName(field);
		if (ivClassName == null) return; // unsupported

		//  private <ivClassName>;   
		//     or, for arrays:
		//  private <ivClassName>[] <javaName>= new <ivClassName>[<fieldName>_num];
		formatter.print("private %% %",
							ivClassName,
							(fieldSpec.isArray()? "[]" : ""),
							javaName);
		if (fieldSpec.isArray()) {
			genArrayFieldInitializer(fieldName, ivClassName);
		}
		formatter.println(";");
	}

	/**
	 * Answer the name of the Java class to use for instance variable
	 * cache for the given Field
	 * @param field Field
	 * @return
	 */
	protected String getIvClassName(Field field) {

		if (field instanceof StringField ) {
			return "String";
			
		} else if (field instanceof IntAccessor ) {
			return "Integer"; 
			
		} else if (field instanceof LongAccessor ) {
			return "Long";
			
		} else if (field instanceof BigDecimalAccessor ) {
			return "BigDecimal";
			
		} else if (field instanceof BigIntegerAccessor ) {
			return "BigInteger";
			
		} else if (field instanceof FloatAccessor ) {
			return "Float";
			
		} else if (field instanceof DoubleAccessor ) {
			return "Double";
			
		} else if (field instanceof ByteArrayField ) {
			return "byte[]";
			
		} else {
			return null; // unsupported
		}
	}
	
	/**
	 * Generate instance variables for the class
	 * @throws IOException
	 */
	protected void genInstanceVariables() throws IOException {
		
		formatter.println();
		formatter.println("protected byte[] _byteBuffer;");
		if (useBufOffset) {
			formatter.println("protected int _byteBufferOffset;");
		}
		formatter.println();
		
		if (!isGenCache()) return;
		formatter.println("// Instance variables used to cache field values");
		
		for (Iterator itr=recordSpec.elementSpecIterator(); itr.hasNext();) {
			ElementSpec spec = (ElementSpec)itr.next();
			if (spec instanceof FieldSpec) {
				FieldSpec fieldSpec = (FieldSpec)spec;
				genInstanceVariable(fieldSpec);
			}
		}
	}

	/**
	 * Generate accessor methods for the given FieldSpec
	 * @param fieldSpec FieldSpec
	 * @throws IOException
	 */
	protected void genSettersAndGetter(FieldSpec fieldSpec) throws IOException {

		Field field = fieldSpec.getField();
		String label = fieldSpec.getLabel();
		if (label == null) return; 
		
		String javaName = getInstanceJavaName(label, fieldSpec.getOperandIndex(), true);
		String argName = getInstanceJavaName(label, fieldSpec.getOperandIndex(), false);
		String ivName = argName;
		String fieldName = getStaticJavaName(label, fieldSpec.getOperandIndex());
		String fieldGetter = "";
		String fieldSetter = "";
		String typeName;
		String typeGetter = "";
		String ivConstructor = "";
		
		if (fieldSpec.isArray()) {
			ivName = ivName + "[_arrayIndex]";
		}

		// write arguments to factory "constructor", by type
		if (field instanceof StringField ) {
			typeName = "String";
			fieldGetter = ".getString";
			fieldSetter = ".putString";
			
		} else if (field instanceof IntAccessor ) {
			typeName = "int"; 
			typeGetter = ".intValue()";
			fieldGetter = ".getInt";
			fieldSetter = ".putInt";
			ivConstructor = "new Integer(";
			
		} else if (field instanceof LongAccessor ) {
			typeName = "long";
			typeGetter = ".longValue()";
			fieldGetter = ".getLong";
			fieldSetter = ".putLong";
			ivConstructor = "new Long(";
			
		} else if (field instanceof FloatAccessor ) {
			typeName = "float";
			typeGetter = ".floatValue()";
			fieldGetter = ".getFloat";
			fieldSetter = ".putFloat";
			ivConstructor = "new Float(";
			
		} else if (field instanceof DoubleAccessor ) {
			typeName = "double";
			typeGetter = ".doubleValue()";
			fieldGetter = ".getDouble";
			fieldSetter = ".putDouble";
			ivConstructor = "new Double(";
			
		} else if (field instanceof ByteArrayField ) {
			typeName = "byte[]";
			fieldGetter = ".getByteArray";
			fieldSetter = ".putByteArray";
			
		} else {  
			return; // unsupported
		}

		// Gen the getter...
		
		formatter.println();
		if (genAccessorJavadoc) {
			genSourceLineComments(fieldSpec);
		}
		//  public <typeName> get<javaName>(<int _arrayIndex>) {
		formatter.println("public % get%(%) {",
							typeName,
							javaName,
							fieldSpec.isArray() ? "int _arrayIndex" : "");
		
		if (isGenCache()) {
			//    if (<ivName == null) {
			formatter.println("if (% == null) {", ivName);
		}
		
		// (for arrays)
		//      int arrayOffset = _arrayIndex * <fieldName>.getByteLength();
		if (fieldSpec.isArray()) {
			formatter.println("int _arrayOffset = _arrayIndex * %.getByteLength();",
								fieldName);
		}
		
		StringBuffer offsets = new StringBuffer();
		if (fieldSpec.getDupCount() > 1) {
			offsets.append(", ");
			if (useBufOffset) {
				offsets.append("_byteBufferOffset + ");
			}
			offsets.append("_arrayOffset");
		} else if (useBufOffset) {
			offsets.append(", _byteBufferOffset");
		}
		if (isGenCache()) {
			//      <ivName = <ivConstructor><fieldName><fieldGetter>(_byteBuffer <,_byteBufferOffset <+ arrayOffset>>)<)>;
			//    }
			formatter.println("% = %%%(_byteBuffer%)%;",
							ivName,
							ivConstructor,
							fieldName,
							fieldGetter,
							offsets,
							ivConstructor.length() > 0 ? ")" : "");
			formatter.println("}");
				
			//    return <ivName><typeGetter>;
			//  }
			formatter.println("return %%;", ivName, typeGetter);
			
		} else {
			///	  return <fieldName>.<fieldGetter>(_byteBuffer <, _byteBufferOffset <, _arrayOffset>>);
			formatter.println("return %%(_byteBuffer%);",
								fieldName,
								fieldGetter,
								offsets);
		}
		formatter.println("}");
		
		//>>>> Gen the setter  <<<<<
		if (!genSetters) return;
		
		formatter.println();
		if (genAccessorJavadoc) {
			genSourceLineComments(fieldSpec);
		}
		//  public void set<javaName>(<int _arrayIndex,> <typeName> <argName>) {
		formatter.println("public void set%(%% %){",
							javaName,
							fieldSpec.isArray() ? "int _arrayIndex, " : "",
							typeName,
							argName);

		if (isGenCache()) {
			//    if (<fieldName>.equals(this.<ivName>, <argName>)) 
			//      return;
			formatter.println("if (%.equals(this.%, %)) {",
								fieldName,
								ivName,
								argName);
			formatter.println("return;");
			formatter.println("}");
		}
		
		// (for arrays)
		//    int arrayOffset = _arrayIndex * <fieldName>.getByteLength();
		if (fieldSpec.isArray()) {
			formatter.println("int _arrayOffset = _arrayIndex * %.getByteLength();",
								fieldName);
		}		
		
		//    <fieldName><fieldSetter>(<argName>, _byteBuffer<,_byteBufferOffset <+ arrayOffset>>);
		formatter.println("%%(%, _byteBuffer%);",
							fieldName,
							fieldSetter,
							argName,
							offsets);
		
		if (isGenCache()) {
			//    this.<ivName> = <ivConstructor><argName><)>;
			formatter.println("this.% = %%%;",
								ivName,
								ivConstructor,
								argName,
								ivConstructor.length() > 0 ? ")" :"");
			
			genClearOverlappingFieldCaches(fieldSpec);
		}
		formatter.println("}");
	}

	/**
	 * Generate the method <code>getByteBuffer()</code> and, if
	 * {@link #isUseBufOffset()} is enabled, the <code>getByteBufferOffset()</code> method.
	 * @throws IOException
	 */
	protected void genBufferAccessors() throws IOException {		

		formatter.println();
		formatter.println("public byte[] getByteBuffer() {");
		formatter.println("return _byteBuffer;");
		formatter.println("}");

		if (useBufOffset) {
			formatter.println();
			formatter.println("public int getByteBufferOffset() {");
			formatter.println("return _byteBufferOffset;");
			formatter.println("}");
		}
	}
	
	/**
	 * Generate code to clear the instance variable caches for any fields
	 * that overlap the given FieldSpec
	 * @param fieldSpec FieldSpec
	 * @throws IOException
	 */
	protected void genClearOverlappingFieldCaches(FieldSpec fieldSpec) throws IOException {
	
		if (fieldSpec.getLabel() == null) return;
		for (Iterator i=recordSpec.overlappingFieldSpecsIterator(fieldSpec); i.hasNext();) {
			FieldSpec oFieldSpec = (FieldSpec)i.next();
			String label = oFieldSpec.getLabel();
			if (label == null) continue;
			String javaName = getInstanceJavaName(label, oFieldSpec.getOperandIndex(), false);
			String fieldName = getStaticJavaName(label, oFieldSpec.getOperandIndex());
			formatter.print("this.%", javaName);
			if (oFieldSpec.isArray()) {
				String ivClassName = getIvClassName(oFieldSpec.getField());
				genArrayFieldInitializer(fieldName, ivClassName);
			} else {
				formatter.print("= null");
			}
			formatter.println(";");
		}
	}

	/**
	 * Generates an initializer value for an array field cache (without trailing semicolon)
     */
	protected void genArrayFieldInitializer(String fieldName, String ivClassName)
			throws IOException {
		if (ivClassName.equals("byte[]")) {
			formatter.print("= new byte[%_num][]", fieldName);				
		} else {
			formatter.print("= new %[%_num]", ivClassName, fieldName);
		}
	}
	
	/**
	 * Generate the field accessor methods for the class
	 * @throws IOException
	 */
	protected void genSettersAndGetters() throws IOException {
		formatter.println();
		
		for (Iterator itr=recordSpec.elementSpecIterator(); itr.hasNext();) {
			ElementSpec spec = (ElementSpec)itr.next();
			if (spec instanceof FieldSpec) {
				FieldSpec fieldSpec = (FieldSpec)spec;
				genSettersAndGetter(fieldSpec);
			}
		}
	}
	
	/**
	 * Generate a java comment containing the source lines associated with the 
	 * given ElementSpec
	 * @param spec ElementSpec
	 * @throws IOException
	 */
	protected void genSourceLineComments(ElementSpec spec) throws IOException {
		
		formatter.println();
		ArrayList lines = spec.getSourceLines();
		if (lines.size() == 0) {
			return; 
		}
		// To avoid printing leading or trailing blank lines,
		// we find the first non-blank line
		int ifirst = 0;
		int ilast = lines.size() -1;
		for (ifirst=0; ifirst <= ilast; ifirst++) {
			String s = (String)lines.get(ifirst);
			if (s.trim().length() > 0) {
				break;
			}
		}
		// Find the last non-blank line
		for (; ilast >= ifirst; ilast--) {
			String s = (String)lines.get(ilast);
			if (s.trim().length() > 0) {
				break;
			}
		}		
		// Print the first non-blank through last non-blank lines
		// wrapped in a "pre" comment.
		for (int i=ifirst; i<= ilast; i++) {
			if (i == ifirst) {
				formatter.println("/** <pre>");
			}
			String s = (String)lines.get(i);
			formatter.print(s);
			if (i < ilast) {
				formatter.println();
			} else {
				formatter.println(" </pre> */");
			}
		}
	}

	/**
	 * Generate Java static variable declarations and initializations for each FieldSpec.
	 * @param javaName the static variable name of the Field
	 * @param fieldSpec the FieldSpec
	 * @throws IOException
	 */
	protected void genStaticFieldDeclaration(String javaName, FieldSpec fieldSpec) throws IOException {

		Field field = fieldSpec.getField();
		String className = getUnqualifiedClassname(field.getClass());
		String access = isGenProtectedFields() ? "protected" : "public";

		if (javaName != null) {
			formatter.print("% static final % % = ", access, className, javaName);
		} else {
			formatter.print("static { ");
		}
		formatter.print("factory.get%(", className);

		// write arguments to factory "constructor", by type
		if (field instanceof StringField ) {
			formatter.print(Integer.toString(field.getByteLength()));
			
		} else if (field instanceof IntAccessor ) {
			IntAccessor binField = (IntAccessor)field;
			formatter.print(Integer.toString(field.getByteLength()));
			formatter.print(", ");
			formatter.print(binField.isSigned() ? "true" : "false" );
			
		} else if (field instanceof LongAccessor ) {
			LongAccessor binField = (LongAccessor)field;
			formatter.print(Integer.toString(field.getByteLength()));
			formatter.print(", ");
			formatter.print(binField.isSigned() ? "true" : "false" );

		} else if (field instanceof FloatAccessor || field instanceof DoubleAccessor) {
			// no factory constructor args required
			
		} else {  // the rest
			formatter.print(Integer.toString(field.getByteLength()));
		}

		if (javaName != null) {
			formatter.println(");");
		} else {
			formatter.println("); } ");
		}

	}
	
	/**
	 * Generate the static Field declarations for the class
	 * @throws IOException
	 */
	protected void genStaticFieldLayout() throws IOException {
		
		String fcn = getUnqualifiedClassname(factory.getClass());
		formatter.println("protected static % factory = new %();", fcn, fcn); 
		formatter.println("static { factory.setStringTrimDefault(%); } ",
							getStringTrimDefault() ? "true" : "false");
		
		for (Iterator itr=recordSpec.elementSpecIterator(); itr.hasNext();) {
			ElementSpec spec = (ElementSpec)itr.next();
			if (spec instanceof FieldSpec) {
				FieldSpec fieldSpec = (FieldSpec)spec;
				genStaticFieldLayout(fieldSpec);
			} else if (spec instanceof ConstantSpec) {
				ConstantSpec constantSpec = (ConstantSpec)spec;
				genConstant(constantSpec);
			}
		}
		formatter.println();
		formatter.println("public static final int _DEFAULT_LEN = factory.getMaximumOffset(); // % bytes",
							"" + factory.getMaximumOffset());
	}
	
	/**
	 * Generate static field declarations for a given FieldSpec
	 * @param fieldSpec the FieldSpec
	 * @throws IOException
	 */
	protected void genStaticFieldLayout(FieldSpec fieldSpec) throws IOException {

		genSourceLineComments(fieldSpec);
		
		// Adjust the current factory offset if it doesn't match the field offset
		int fieldOffset = fieldSpec.getField().getOffset();
		if (factory.getOffset() != fieldOffset) {
			formatter.println("static { factory.incrementOffset(%); } ",
								Integer.toString(fieldOffset - factory.getOffset()));
			factory.setOffset(fieldOffset);
		}
		
		// If the field dup factory is 0, the push the current factory offset 
		if (fieldSpec.getDupCount() == 0) {
			formatter.println("static { factory.pushOffset(); } ");
			factory.pushOffset();
		}

		Field field = fieldSpec.getField();
		factory.setOffset(factory.getOffset() + field.getByteLength());
		
		String javaName = null;
		if (fieldSpec.getLabel() != null) {
			javaName = getStaticJavaName(fieldSpec.getLabel(), fieldSpec.getOperandIndex());
		}
		genStaticFieldDeclaration(javaName, fieldSpec);
		
		// If the field dup factor is 0, the pop the current factory offset 
		if (fieldSpec.getDupCount() == 0) {
			formatter.println("static { factory.popOffset(); } ");
			factory.popOffset();
			
		// If the field dup factor is > 1, then reserve space for the array
		} else if (fieldSpec.isArray()) {

			String nen = Integer.toString(fieldSpec.getDupCount());
			String access = isGenProtectedFields() ? "protected" : "public";
			if (javaName != null) {
				nen = javaName + "_num";
				formatter.println("% static final int % = %;",
									access,
									nen,
									Integer.toString(fieldSpec.getDupCount()));
			}
			
			formatter.println("static { factory.incrementOffset((% - 1) * %); } ",
								nen,
								javaName != null 
									? "" + javaName + ".getByteLength()"
									: Integer.toString(field.getByteLength()));
												
			factory.incrementOffset((fieldSpec.getDupCount()-1) * field.getByteLength());
		}

	}
	
	/**
	 * Generate a file trailer
	 * @throws IOException
	 */
	protected void genTrailer() throws IOException {

		formatter.println("}");
		formatter.println();	
	}

	/**
	 * Static helper method to return the simple/unqualified name of a java Class
	 * @param c the Class
	 * @return String the simple class name
	 */
	private static String getUnqualifiedClassname(Class c) {
		String fqn = c.getName();
		int i = fqn.lastIndexOf('.');
		if (i < 0) return fqn;
		return fqn.substring(i+1);
	}
	
	/**
	 * Answer a camel-cased Java name used for instance variables
	 * @param label the Assembler label
	 * @param index the operand index
	 * @param upperFirst boolean if true, the first letter is uppercased
	 * @return String
	 */
	protected String getInstanceJavaName(String label, int index, boolean upperFirst ) {
		StringTokenizer strtok = new StringTokenizer(label, "_");
		StringBuffer sb = new StringBuffer();
		while (strtok.hasMoreTokens()) {
			String tok = strtok.nextToken();
			char c = tok.charAt(0);
			if (!upperFirst) {
				sb.append(Character.toLowerCase(c));
				upperFirst = true;
			} else {
				sb.append(Character.toUpperCase(c));
			}
			char prevChar = c;
			for (int i=1; i<tok.length(); i++) {
				c = tok.charAt(i);
				if (!isCamelCaseAccessors()) {
					sb.append(c);
				} else if (isDigitOrHashOrAt(prevChar) && !isDigitOrHashOrAt(c)) {
					sb.append(Character.toUpperCase(c));
				} else {
					sb.append(Character.toLowerCase(c));
				}
				prevChar = c;
			}
		}
		if (index > 0) {
			sb.append("_");
			sb.append(Integer.toString(index));
		}
		String answer = sb.toString();
		answer = toJavaIdentifer(answer);		
		return answer;
	}
	
	/**
	 * Answer a name to be used as a Java static field variable name
	 * @param label the Assembler label
	 * @param index the operand index
	 * @return String the Java variable name
	 */
	protected String getStaticJavaName(String label, int index ) {
		label = toJavaIdentifer(label);
		StringTokenizer strtok = new StringTokenizer(label, "_");
		StringBuffer sb = new StringBuffer();
		while (strtok.hasMoreTokens()) {
			String tok = strtok.nextToken();
			sb.append(tok.toUpperCase());
			if (strtok.hasMoreTokens()) {
				sb.append("_");
			}
		}
		if (index > 0) {
			sb.append("_");
			sb.append(Integer.toString(index));
		}
		return sb.toString();
	}
	
	/**
	 * Answer a String that maps assembler labels into valid
	 * Java identifier characters.  '#' is replaced by "_hash_", 
	 * and '@' is replaced by _at_.
	 */
	protected String toJavaIdentifer(String label) {
		StringBuffer sb = new StringBuffer();
		int last = label.length()-1;
		char prev = '\0';
		for (int i=0; i<=last; i++) {
			char c = label.charAt(i);
			if (c == '#') {
				if (i > 0 && prev != '_') sb.append('_');
				sb.append("hash");
				if (i != last) sb.append('_');
			} else if (c == '@') {
				if (i > 0 && prev != '_') sb.append('_');
				sb.append("at");
				if (i != last) sb.append('_');
			} else {
				sb.append(c);
			}
			prev = c;
		}
		return sb.toString();
	}

	private boolean isDigitOrHashOrAt(char c) {
		return Character.isDigit(c)
			|| c == '#'
			|| c == '@';
	}
	
}
