/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;

import com.ibm.jzos.fields.AssemblerDatatypeFactory;
import com.ibm.jzos.fields.Field;
import com.ibm.jzos.fields.IbmDoubleField;
import com.ibm.jzos.fields.IbmFloatField;
import com.ibm.jzos.recordgen.ADataRecord;
import com.ibm.jzos.recordgen.ADataRecordReader;

/**
 * A class that converts a AData file, output from the IBM High Level Assembler,
 * into an abstract {@link RecordSpec} object model.
 * <p>
 * This class is used by {@link RecordClassGenerator} to generate Java source
 * code from Assembler AData files.
 * <p>
 * @see ADataRecord
 */
public class RecordSpecBuilder {

	private RecordSpec recordSpec;
	private ArrayList sourceRecords;
	private int prevRecordIdx = 0;
	private int curRecordIdx = 0;

	
	private String selectedSectionName = null;
	private boolean dumpSourceRecords = false;
	private boolean commentsPrecedeDeclaration = false;
	private int selectedSectionId = 0;
	
	/**
	 * Process an Assembler AData file.
	 * @param filename the name of the ADATA file
	 * @return RecordSpec
	 * @throws IOException
	 */
	public RecordSpec process(String filename) throws IOException {
		
		recordSpec = new RecordSpec();
		
		ADataRecordReader reader = new ADataRecordReader(filename);
		sourceRecords = readSourceRecords(reader);

		if (dumpSourceRecords) dumpSourceRecords(sourceRecords);
		
		for (int i=1; i<sourceRecords.size(); i++) { // skip first dummy object
			ADataSourceRecord rec = (ADataSourceRecord)sourceRecords.get(i);
			
			if (rec.getDCDSRecord() != null) {
				String label = null;
				if (rec.getSymbolRecord() != null) {
					ADataSymbolRecord symRec = rec.getSymbolRecord();
					label = symRec.getSymbolName();
				}
				curRecordIdx = i;
				processDCDSRecord(rec.getDCDSRecord(), label);
			} else if (rec.getSymbolRecord() != null) {
				curRecordIdx = i;
				processSymbolRecord(rec.getSymbolRecord());
			}
		}
		return recordSpec;
	}

	/**
	 * Process an Assembler AData file and return the DSECT names included.
	 */
	public static String[] getDsectNames(String filename) throws IOException {

		RecordSpecBuilder instance = new RecordSpecBuilder();
		List answer = new ArrayList();
		ADataRecordReader reader = new ADataRecordReader(filename);
		ArrayList sourceRecords = instance.readSourceRecords(reader);

		for (int i=1; i<sourceRecords.size(); i++) { // skip first dummy object
			ADataSourceRecord rec = (ADataSourceRecord)sourceRecords.get(i);
			ADataSymbolRecord symrec = null;
			if ((symrec = rec.getSymbolRecord()) != null) {
				int type = symrec.getTYPE();
				if (type == ADataSymbolRecord.ADSYM_TYPE_DSECT) {
					answer.add(symrec.getSymbolName());
				}
			}
		}
		return (String[])answer.toArray(new String[answer.size()]);
		
	}
	
	/**
	 * Process a record produced by a DC/DS statement, attached to the given label name
	 * and preceded by given sourceLines.
	 */
	private void processDCDSRecord(ADataDCDSRecord rec, String label) {

		// if we haven't found our section, or this isn't it then ignore the field
		int esdid = rec.getESDID();
		if (selectedSectionId == 0 || selectedSectionId != esdid) {
			return;  
		}
		
		if (rec.getTYPE_FLAG() != ADataDCDSRecord.ADDCDS_TYPE_DC 
				&& rec.getTYPE_FLAG() != ADataDCDSRecord.ADDCDS_TYPE_DS) {
			return;   // we don't support CXD, DXD, CCW, etc.  
		}
		
		int sourceStmt = rec.getSTMT();
		
		ADataDCDSOperand operand = rec.getFirstOperand();
		int opsNum = rec.getOPS_NUM();
		for (int i=0; operand != null && i<opsNum; i++) {
			processDCDSOperand(operand, i, label, sourceStmt, i == 0);
			operand = operand.getNextOperand();
		}
	}

	/**
	 * Process one operand from a DC/DS record
	 */
	private void processDCDSOperand(ADataDCDSOperand operand, int operandIndex,
							String label, int sourceStmt, boolean firstOperand) {
	
		int dup = operand.getDUP();
		char type = operand.getTYPE_ATTR().charAt(0);
		char ext = operand.getTYPE_EXT().charAt(0);
		ADataDCDSValue aDataValue = operand.getFirstValue();
		int byteLen = aDataValue.getBYTE_LEN();
		int offset = operand.getLOCTR();
		int bitOffset = operand.getBIT_OFFSET();

		Field field = null;
		if (bitOffset == 0) {
			field = makeField(type, ext, offset, byteLen);
		}
		
		if (field != null) {
			FieldSpec fieldSpec = new FieldSpec();
			fieldSpec.setSourceStmt(sourceStmt);
			if (firstOperand) { // only add source lines for first operand
				fieldSpec.setSourceLines(getSourceLines());
			}
			fieldSpec.setLabel(label);
			fieldSpec.setField(field);
			fieldSpec.setDupCount(dup);
			fieldSpec.setOperandIndex(operandIndex);
			// stuff just for complete XML generation:
			fieldSpec.setBitOffset(operand.getBIT_OFFSET());
			fieldSpec.setTypeAttr(operand.getTYPE_ATTR());
			if (ext != ' ' && ext != '\0') 
				fieldSpec.setTypeExt(operand.getTYPE_EXT());
			recordSpec.addElementSpec(fieldSpec);
		}
	}
	
	public static Field makeField(char type, char ext, int offset, int byteLen) {
		
		AssemblerDatatypeFactory factory = new AssemblerDatatypeFactory();
		factory.setOffset(offset);
		
		Field field = null;
		
		switch (type) {
		case 'C':
			field = factory.getStringField(byteLen);
			break;
			
		case 'A':
		case 'F':
		case 'H':
		case 'R':
		case 'S':
		case 'V':
		case 'Y':
			field = factory.getBinaryField(byteLen, true);
			break;
			
		case 'X':
		case 'B':
			if (byteLen > 0 && byteLen <= 4) {
				field = factory.getBinaryField(byteLen, false);
			} else {
				field = factory.getByteArrayField(byteLen);
			}
			break;
			
		case 'E':
		case 'D':
		case 'K':
			if (ext != 'B' && byteLen == IbmFloatField.BYTE_LENGTH) {
				field = factory.getIbmFloatField();
			} else if (ext != 'B' && byteLen == IbmDoubleField.BYTE_LENGTH) {
				field = factory.getIbmDoubleField();
			} else {
				field = factory.getByteArrayField(byteLen);  // not supported 
			}
			break;
						
		case 'P':
			field = factory.getPackedDecimalField(byteLen, 0, true);
			break;
			
		case 'Z':
			field = factory.getZonedDecimalField(byteLen, 0, true);
			break;
			
		default:
			field = factory.getByteArrayField(byteLen);
		}
		return field;
				
	}
	/**
	 * Process a Symbol record from the ADATA file.
	 * We use CSECT, DSECT, and EQU symbols only to generate ConstantSpecs.
	 * @param rec the Symbol record
	 * @param sourceLines the source lines that preceded the record
	 */
	private void processSymbolRecord(ADataSymbolRecord rec) {
		
		int type = rec.getTYPE();
		String label = rec.getSymbolName();
		int esdid = rec.getESDID();
		
		
		// If for a CSECT or DSECT
		if (type == ADataSymbolRecord.ADSYM_TYPE_SECTION 
				|| type == ADataSymbolRecord.ADSYM_TYPE_DSECT) {
			
			if (selectedSectionId == 0) {
				// we haven't found our section.
				// if we are waiting for a specific name  or this is it, then
				// remember this sectionId as the one we want
				if (selectedSectionName == null 
					 || selectedSectionName.equalsIgnoreCase(label)) {
					selectedSectionId = esdid;
					recordSpec.setSectionName(label);
				} else {
					return;
				}
			} else {
				// we have found our section
				// if this isn't it, ignore the symbol
				if (selectedSectionId != esdid) return;
			}
		
		} else if (type != ADataSymbolRecord.ADSYM_TYPE_EQU ) {
			return;  // bail if not CSECT, DSECT, or EQU
		}

		// If we haven't found our section or this isn't it then bail
		if (selectedSectionId == 0 || selectedSectionId != esdid) {
			return;  
		}

		ConstantSpec constantSpec = new ConstantSpec();
		constantSpec.setSourceStmt(rec.getSTMT());
		constantSpec.setSourceLines(getSourceLines());
		constantSpec.setLabel(label);
		constantSpec.setRelocatable((rec.getRELOC_FLAG() & ADataSymbolRecord.ADSYM_RELOC) != 0);
		
		// stuff we keep for Xml writer
		constantSpec.setTypeAttr(rec.getATTR());
		
		if (rec.getASM_TYPE().equals("C")) {
			byte[] b = new byte[1];
			b[0] = (byte)rec.getLOCTR();
			String s;
			try {
				s = new String(b, 0, 1, "Cp1047");
			} catch (UnsupportedEncodingException uce) {
				throw new RuntimeException(uce);
			}
			constantSpec.setValue(new Character(s.charAt(0)));
		} else {
			constantSpec.setValue(new Integer(rec.getLOCTR()));
		}
		recordSpec.addElementSpec(constantSpec);
	}
	
	/**
	 * Process the SOURCE records from an {@link ADataRecordReader) and
	 * return them in a list.  As DCDS or Symbol records are read, they
	 * are associated with the preceding source record(s).
	 * @param reader
	 * @return ArrayList of ADataSourceRecords
	 * @throws IOException
	 */
	private ArrayList readSourceRecords(ADataRecordReader reader) throws IOException {
		ArrayList sourceRecords = new ArrayList();
		sourceRecords.add(new Object());  // dummy for zero-index

		ADataRecord rec = null;
		ADataSourceRecord sourceRec = null;
		ADataSymbolRecord symbolRec = null;
		while ((rec = reader.readRecord()) != null) {
				
			if (rec.getRECTYPE() == ADataSourceRecord.ADATA_RECSOURCE) {
				ADataSourceRecord nextSourceRec = new ADataSourceRecord(rec.getBytes());
				if (sourceRec == null || 
					nextSourceRec.getSTMT() > sourceRec.getSTMT()) {
					sourceRec = nextSourceRec;
					sourceRecords.add(sourceRec);
				} else {
					// Add the nextSourceRec as the last continuation record in the previous
					// sourceRec's chain
					ADataSourceRecord contRec = sourceRec;
					while (contRec.getContinuationRecord() != null) {
						contRec = contRec.getContinuationRecord();
					}
					contRec.setContinuationRecord(nextSourceRec);
				}
			} else if (rec.getRECTYPE() == ADataDCDSRecord.ADATA_RECDCDS) {
				// ByteUtil.dumpHex("ADataDCDSRecord", rec.getBytes(), System.err); // DEBUG
				sourceRec.setDCDSRecord(new ADataDCDSRecord(rec.getBytes()));	
			} else if (rec.getRECTYPE() == ADataSymbolRecord.ADATA_RECSYM) {
				// ByteUtil.dumpHex("ADataSymbolRecord", rec.getBytes(), System.err); // DEBUG
				symbolRec = new ADataSymbolRecord(rec.getBytes());
				sourceRec = (ADataSourceRecord)sourceRecords.get(symbolRec.getSTMT());
				sourceRec.setSymbolRecord(symbolRec);
			}
		}
		return sourceRecords;
	}

	/**
	 * Get the source lines associated with the record (located at curRecordIdx in
	 * sourceRecords).  If commentsPrecedeDeclaration == true, take the source lines
	 * preceding the current record, up to the previous record.
	 * Otherwise, take the source lines following the current record until a new
	 * symbol or DCDS record is encountered.
	 */
	private ArrayList getSourceLines() {
		ArrayList sourceLines = new ArrayList();
		ADataSourceRecord sourceRec;
		if (commentsPrecedeDeclaration) {
			if (prevRecordIdx == 0) {
				//For DSECT, just take the current source line, skipping any preceding lines
				sourceRec = (ADataSourceRecord)sourceRecords.get(curRecordIdx);
				sourceLines.addAll(getSourceLines(sourceRec));
			} else {
				for (int i = curRecordIdx; i > prevRecordIdx; i--) {
					sourceRec = (ADataSourceRecord)sourceRecords.get(i);
					sourceLines.addAll(0, getSourceLines(sourceRec));
				}
			}
			prevRecordIdx = curRecordIdx;
		} else {
			sourceRec = (ADataSourceRecord)sourceRecords.get(curRecordIdx);
			sourceLines.addAll(getSourceLines(sourceRec));
			for (int i = curRecordIdx + 1; i < sourceRecords.size(); i++) {
				sourceRec = (ADataSourceRecord)sourceRecords.get(i);
				if (sourceRec.getDCDSRecord() == null && sourceRec.getSymbolRecord() == null) {
					sourceLines.addAll(getSourceLines(sourceRec));
				} else {
					break;
				}
			}
		}
		return sourceLines;
	}

	private ArrayList getSourceLines(ADataSourceRecord rec) {
		ArrayList sourceLines = new ArrayList();
		sourceLines.add(scrubSourceLine(rec.getSourceLine()));
		for (ADataSourceRecord crec = rec.getContinuationRecord(); 
				crec != null;
				crec = crec.getContinuationRecord()) {
			sourceLines.add(scrubSourceLine(crec.getSourceLine()));
		}
		return sourceLines;
	}

	/**
	 * For debugging purposes, dump the list of SourceRecords on System.err
	 */
	private void dumpSourceRecords(ArrayList sourceRecords) throws IOException {
		PrintWriter wtr = new PrintWriter(System.err);		
		try {
		for (int i=1; i<sourceRecords.size(); i++) {
			ADataSourceRecord rec = (ADataSourceRecord)sourceRecords.get(i);
			rec.dumpOn(wtr);
			for (ADataSourceRecord crec = rec.getContinuationRecord();
					crec != null;
					crec = crec.getContinuationRecord()) {
				crec.dumpOn(wtr);
			}
			if (rec.getSymbolRecord() != null) {
				rec.getSymbolRecord().dumpOn(wtr);
			}
			if (rec.getDCDSRecord() != null) {
				rec.getDCDSRecord().dumpOn(wtr);
			}
		}
		} finally {
			wtr.flush();
		}
	}
	
	/**
	 * Trim a source line at column 72, replace any Java comments and
	 * replace literal @ signs with the unicode sequence.
	 */
	private static String scrubSourceLine(String sourceLine) {
		if (sourceLine == null) return null;
		if (sourceLine.length() > 72) {
			sourceLine = sourceLine.substring(0,72);
		}
		sourceLine = sourceLine.replaceAll("(\\/\\*)|(\\*\\/)", "**");
		sourceLine = sourceLine.replaceAll("@", "&#x0040;");
		sourceLine = sourceLine.replaceAll("\\s+$", "");  // trim trailing white space
		return sourceLine;
	}

	/**
	 * Answer the name of the DSECT or CSECT that is to be scavenged.
	 * @return String the DSECT or CSECT name
	 */
	public String getSelectedSectionName() {
		return selectedSectionName;
	}

	/**
	 * @see #getSelectedSectionName()
	 */
	public void setSelectedSectionName(String selectedSectionName) {
		this.selectedSectionName = selectedSectionName;
	}

	/**
	 * Sets a debug flag which causes the AData source records to be dumped to STDERR
	 * @param dumpSourceRecords
	 */
	public void setDumpSourceRecords(boolean dumpSourceRecords) {
		this.dumpSourceRecords = dumpSourceRecords;
	}

	/**
	 * Sets a flag to indicate how to generate source comments from the ADATA DSECT.
	 * @param commentsPrecedeDeclaration if true, comments are assumed to precede their
	 * associated declaration.
	 */
	public void setCommentsPrecedeDeclaration(boolean commentsPrecedeDeclaration) {
		this.commentsPrecedeDeclaration = commentsPrecedeDeclaration;
	}
}
