/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

/**
 * Base class for the abstract spec tree used to represent a COBOL copybook
 * built from the compiler's ADATA file.
 * @see RecordClassGenerator
 * @see RecordSpecTreeBuilder
 */
public abstract class ElementSpec {

	/**
	 * Static helper to get a simple class name for a Class
	 * @param c Class
	 * @return String
	 */
	static String getUnqualifiedClassname(Class c) {
		String fqn = c.getName();
		int i = fqn.lastIndexOf('.');
		if (i < 0) return fqn;
		return fqn.substring(i+1);
	}	
	private int level;
	private Offset offset; 
	private ElementSpec parent;	
	private int size;	
	private ArrayList<String> sourceLines = new ArrayList<String>();
	private int sourceStmt = 0;
	private String symbol;
	private boolean symbolUnique = true;
	private int symbolId;

	private boolean varLoc;
	
	protected static final String NL = System.getProperty("line.separator"); 
	protected static final String INDENT = "   ";

	/**
	 * Dump debug information for the receiver
	 * <p>
	 * @param writer
	 * @throws IOException
	 */
	public void dumpOn(Writer writer) throws IOException {
		String pindent = getIndentString('.');
		writer.write("\n\n");
		for (Iterator i=getSourceLines().iterator(); i.hasNext();) {
			writer.write((String)i.next());
			writer.write("\n");
		}
		writer.write(pindent);
		writer.write(toString());
		writer.write("\n");
		
	}

	/**
	 * Return the grouping (nesting) level of the element.
	 */
	public int getLevel() {
		return level;
	}

	/**
	  * Return the offset from the beginning of the record.  If this is a
	  * variable located item (i.e. element follows and ODO), this will be
	  * a VariableOffset.
	  */
	public Offset getOffset() {
		return offset;
	}

	/**
	 * For group(ed) items, the containing group.  Null if this is the top level group.
	 * @return boolean
	 */
	public ElementSpec getParent() {
		return parent;
	}

	/**
	 * Return the size of the element.  If the element has an OCCURS clause, this
	 * is the size of a single occurence.  If this is a Group element containing
	 * an ODO in its subtree, a value of -1 is returned.
	 */
	public int getSize() {
		return size;
	}

	/**
	 * Return a List of Strings which are typically Assembler comments for the receiver's element.
	 * @return ArrayList
	 */
	public ArrayList<String> getSourceLines() {
		return sourceLines;
	}
	
	/**
	 * Return the source lines as a line terminated String
	 * @since 2.4.4
	 */
	public String getSourceLinesAsString() {
		StringBuffer sb = new StringBuffer();
		for (Iterator<String> iter=sourceLines.iterator(); iter.hasNext();) {
			String line = iter.next();
			sb.append(line);
			if (iter.hasNext()) sb.append(NL);
		}
		return sb.toString();
	}

	/**
	 * Return an integer which is the COBOL source statement number
	 * @return int
	 */
	public int getSourceStmt() {
		return sourceStmt;
	}
	
	/**
	 * Return a String which is the COBOL symbol for the element
	 * @return String
	 */
	public String getSymbol() {
		return symbol;
	}
	
	/**
	 * Return true if the name is unique in the compilation unit.  If false, 
	 * it may need to be qualified.
	 * @return boolean
	 */
	public boolean isSymbolUnique() {
		return symbolUnique;
	}

	/**
	 * Return true if the element is variably located (it follows an ODO element).
	 */
	public boolean isVarLoc() {
		return varLoc;
	}

	/** 
	 * @see #getLevel()
	 */
	public void setLevel(int level) {
		this.level = level;
	}

	/**
	 * @see #getOffset()
	 */
	public void setOffset(Offset offset) {
		this.offset = offset;
	}

	/**
	 * @see #getParent()
	 */
	public void setParent(ElementSpec parent) {
		this.parent = parent;
	}

	/**
	 * @see #getSize()
	 */
	public void setSize(int size) {
		this.size = size;
	}

	/**
	 * @see #getSourceLines()
	 * @since 2.4.4
	 */
	public void setSourceLines(ArrayList<String> lines) {

		// Find the first non-blank line
		int ifirst = 0;
		int ilast = lines.size() -1;
		for (ifirst=0; ifirst <= ilast; ifirst++) {
			String s = lines.get(ifirst);
			if (s.trim().length() > 0) {
				break;
			}
		}
		// Find the last non-blank line
		for (; ilast >= ifirst; ilast--) {
			String s = lines.get(ilast);
			if (s.trim().length() > 0) {
				break;
			}
		}	
		// Copy the non-blank lines into a new ArrayList, trimming each
		this.sourceLines = new ArrayList<String>(ilast-ifirst+1);
		for (int i=ifirst; i<= ilast; i++) {
			String s = lines.get(i);
			s = s.replaceAll("\\s+$", "");  // trim trailing white space
			this.sourceLines.add(s);
		}
		
	}
	
	/**
	 * @set #getSourceLines()
	 * @since 2.4.4
	 */
	public void setSourceLines(String s) {
		ArrayList<String> lines = new ArrayList<String>();
		String line;
		BufferedReader sr = new BufferedReader(new StringReader(s));
		try {
			while ((line = sr.readLine()) != null)	{
				lines.add(line);
			}
		} catch (IOException canthappen) {
			throw new RuntimeException(canthappen);
		}
		setSourceLines(lines);		
	}

	/**
	 * @see #getSourceStmt()
	 */
	public void setSourceStmt(int sourceStmt) {
		this.sourceStmt = sourceStmt;
	}

	/**
	 * @see #getSymbol()
	 */
	public void setSymbol(String label) {
		this.symbol = label;
	}

	/**
	 * @see #isSymbolUnique()
	 */
	public void setSymbolUnique(boolean symbolUnique) {
		this.symbolUnique = symbolUnique;
	}

	
	/**
	 * @return the unique symbol record id from the adata symbol record
	 * @since 2.4.4
	 */
	public int getSymbolId() {
		return symbolId;
	}

	/**
	 * @see #getSymbolId()
	 * @since 2.4.4
	 */
	public void setSymbolId(int symbolId) {
		this.symbolId = symbolId;
	}

	/**
	 * @see #isVarLoc()
	 */
	public void setVarLoc(boolean varLoc) {
		this.varLoc = varLoc;
	}

	/**
	 * A debug representation of the receiver.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer(getSymbol());
		sb.append(" ");
		String name = getClass().toString();
		name = name.substring(name.lastIndexOf('.')+1);
		sb.append(name);
		sb.append(" (");
		sb.append(getOffset());
		sb.append(",");
		sb.append(getSize());
		sb.append(")");
		return sb.toString();		
	}

	protected String getIndentString(char c) {
		StringBuffer sb = new StringBuffer();
		for (int i=0; i<level; i++) {
			sb.append(c);
		}
		return sb.toString();
	}
	
	/*--- methods relating to XML parsing and generation -----*/
	
	/**
	 * Write the receiver in XML form
	 * 
	 * @param writer XMLStreaWriter
	 * @param indentLevel int the indenting level for quasi-pretty printing the XML
	 * @since 2.4.4
	 */
	public abstract void writeXmlOn(XMLStreamWriter writer, int indentLevel) throws XMLStreamException;
	
	/**
	 * Write ElementSpec-level attributes in XML form
	 * @since 2.4.4
	 */
	protected void writeXmlAttributesOn(XMLStreamWriter writer) throws XMLStreamException {
		writer.writeAttribute("symbol", getSymbol());
		writer.writeAttribute("level", Integer.toString(getLevel()));
		writer.writeAttribute("symbolId", Integer.toString(getSymbolId()));
		if (getSize() >= 0) {
			writer.writeAttribute("size", Integer.toString(getSize()));
		}
		writer.writeAttribute("sourceStmt", Integer.toString(getSourceStmt()));
		if (offset != null) {
			offset.writeXmlAttributesOn(writer);
		}
		if (!isSymbolUnique()) {
			writer.writeAttribute("symbolUnique", "false");
		}
		if (isVarLoc()) {
			writer.writeAttribute("varLoc", "true");
		}

	}
	
	/**
	 * Write source lines in XML form
	 * @since 2.4.4
	 */
	protected void writeXmlSourceLinesOn(XMLStreamWriter writer) throws XMLStreamException {
		writer.writeCharacters(NL);
		writer.writeCharacters(getSourceLinesAsString());
	}

	/**
	 * Helper for writing a newline during XML rendering
	 * @since 2.4.4 
	 */
	protected void writeXmlNewline(XMLStreamWriter writer, int indentLevel) throws XMLStreamException {
		writer.writeCharacters(NL);
		for (int i=indentLevel; i>0; i--) {
			writer.writeCharacters(INDENT);
		}
	}

	/**
	 * Parse the receiver from XML
	 * @since 2.4.4
	 */
	protected abstract void parseXmlFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException;
	
	/**
	 * Parse ElementSpec level attributes from an XML parser
	 * @since 2.4.4
	 */
	protected void parseXmlAttributesFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		setSymbol(parser.getAttributeValue("symbol"));
		setLevel(parser.getAttributeIntValue("level"));
		setSymbolId(parser.getAttributeIntValue("symbolId"));
		setSize(parser.getAttributeIntValue("size"));
		setSourceStmt(parser.getAttributeIntValue("sourceStmt"));
		setOffset(Offset.parseXmlFrom(parser));
		setSymbolUnique(parser.getAttributeBooleanValue("symbolUnique", true));
		setVarLoc(parser.getAttributeBooleanValue("varLoc", false));
	}
	
	
	/**
	 * Parses source lines, up to and including the enclosing element end.
	 * <b/>
	 * This would be used for ElementSpecs that do not have children other than source lines
	 * (elementSpecs that are not StorageSpecs).
	 * @since 2.4.4 
	 */
	protected ArrayList<String> parseXmlSourceLinesFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		XMLStreamReader reader = parser.getReader();
		
		// we start with a buffer that contains any existing source data, which might have been
		// parsed when parsing StorageSpec children
		StringBuffer sb = new StringBuffer(getSourceLinesAsString()); 
		boolean endElement = false;
		while (!endElement && reader.hasNext()) {
			switch (reader.next()) {
			case XMLStreamConstants.END_ELEMENT:
				endElement = true;
				break;
			case XMLStreamConstants.CHARACTERS:
				sb.append(new String(reader.getTextCharacters(),
										reader.getTextStart(),
										reader.getTextLength()));
				break;
			case XMLStreamConstants.CDATA:
				if (reader.hasText()) {
					sb.append(reader.getText());
				}
				break;
			}
		}
		setSourceLines(sb.toString());
		return sourceLines;
	}
	
}
