/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import java.util.Iterator;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import com.ibm.jzos.fields.BigDecimalAccessor;
import com.ibm.jzos.fields.BigIntegerAccessor;
import com.ibm.jzos.fields.Field;

/**
 * An object to describe a group in a COBOL COPYBOOK.
 */
public class GroupSpec extends StorageSpec {
	private boolean containsOdo = false;
	
	static final String XML_ELEMENT_NAME = "group";

	/**
	 * Answer true if the receiver contains any fields that will require java.Math
	 */
	public boolean containsBigDecimalsOrIntegers() {
		for (Iterator<ElementSpec> i=getChildren().iterator(); i.hasNext();) {
			ElementSpec child = i.next();
			if (child instanceof GroupSpec) {
				GroupSpec groupSpec = (GroupSpec)child;
				if (groupSpec.containsBigDecimalsOrIntegers()) return true;
			} else if (child instanceof FieldSpec) {
				FieldSpec fieldSpec = (FieldSpec)child;
				Field field = fieldSpec.getField();
				if (field instanceof BigDecimalAccessor || field instanceof BigIntegerAccessor) 
					return true;
			}
		}
		return false;
	}

	/**
	 * Answer true if the receiver contains an "OCCURS DEPENDING ON" clause.
	 */
	public boolean containsOdo() {
		return containsOdo;
	}

	/** 
	 * Answer true if this storage spec is or contains a field
	 * that is the object of a "OCCURS DEPENDS ON" clause.
	 * In other words, it contains the size of an ODO.
	 */
	public boolean containsOdoObject() {
		for (Iterator<ElementSpec> i=getChildren().iterator(); i.hasNext();) {
			ElementSpec child = i.next();
			if (child instanceof StorageSpec) {
				StorageSpec storageSpec = (StorageSpec)child;
				if (!storageSpec.containsOdoObject()) return false;
			}
		}
		return true;
	}

	/**
	 * Answer the size, in bytes, of the storage described by this GroupSpec.
	 * If the GroupSpec contains an ODO, then answer -1, indicating that 
	 * the size is not fixed.
	 */
	public int getSize() {
		return containsOdo() ? -1 : super.getSize();
	}
	
	/**
	 * Answer true if cached getter/setter methods should be generated
	 * for the receiver.  Overridden so that the root-level GroupSpec
	 * will never get cached accessors.
	 */
	public boolean isGenerateCachedAccessors() {
		return super.isGenerateCachedAccessors()
		     && getParent() != null;
	}
	
	/**
	 * Answer true if a StringField overlay should be generated.
	 * <strong>Note: this feature is disabled, so always answer false.</strong>
	 * <p>
	 * This might otherwise be true if the following conditions are all met:
	 * <ul>
	 * <li>It is not an array (OCCURS or ODO)</li>
	 * <li>It doesn't contain an ODO</li>
	 * <li>It only contains alpha-numeric fields</li>
	 * </li>
	 */
	public boolean isGenerateStringField() {
		return false;
	}

	/**
	 * Set whether the receiver contains an "OCCURS DEPENDING ON" clause.
	 */
	public void setContainsOdo(boolean containsOdo) {
		this.containsOdo = containsOdo;
	}
	
	public void setGenerateCachedAccessors(boolean generateCachedAccessors) {
		super.setGenerateCachedAccessors(generateCachedAccessors);
		for (Iterator<ElementSpec> i=getChildren().iterator(); i.hasNext();) {
			ElementSpec child = i.next();
			if (child instanceof StorageSpec) {
				StorageSpec storageSpec = (StorageSpec)child;
				storageSpec.setGenerateCachedAccessors(generateCachedAccessors);
			}
		}
	}	
	
	/**
	 * @see ElementSpec#writeXmlOn(XMLStreamWriter, int)
	 */
	@Override
	public void writeXmlOn(XMLStreamWriter writer, int indentLevel)
			throws XMLStreamException {
		writeXmlNewline(writer, indentLevel);
		writer.writeStartElement(XML_ELEMENT_NAME);
		writeXmlAttributesOn(writer);
		writeXmlSourceLinesOn(writer);
		writeXmlChildrenOn(writer, indentLevel+1);
		writeXmlNewline(writer, indentLevel);
		writer.writeEndElement();
	}
		
}
