/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import com.ibm.jzos.fields.BigDecimalAccessor;
import com.ibm.jzos.fields.BigIntegerAccessor;
import com.ibm.jzos.fields.DoubleAccessor;
import com.ibm.jzos.fields.Field;
import com.ibm.jzos.fields.FloatAccessor;
import com.ibm.jzos.fields.IntAccessor;
import com.ibm.jzos.fields.LongAccessor;
import com.ibm.jzos.fields.StringField;

/**
 * A class with a static interface used in generating
 * code for Cobol literals which appear both in ConditionSpecs
 * and as VALUE clauses in StorageSpecs.
 */
public class Literal {

	public static final char DOUBLE_QUOTE = '\"';
	public static final char SINGLE_QUOTE = '\'';
	
	/**
	 * Answer a String containing the Java code used to
	 * create a Java object representing the Cobol 
	 * literal specified in the given valueClause.
	 * Answers null if the given valueClause is not supported.
	 * 
	 * @param valueClause the Cobol valueClause obtained from the SymbolRecord
	 * @param field the Field that is to be initialized.
	 * @return String the Java code used to create a literal or null if not supported
	 */
	public static String toJavaCode(String valueClause, Field field) {
		
		if (valueClause == null || valueClause.length() == 0) return ""; // huh?
		
		char first = valueClause.charAt(0);
		if (first == SINGLE_QUOTE || first == DOUBLE_QUOTE) {
			return fromLiteralString(valueClause);
		} else if (Character.isLetter(first)) {
			return fromFigurativeLiteral(valueClause, field);
		} else if (Character.isDigit(first) || first == '+' || first == '-') {
			return fromNumericLiteral(valueClause, field);
		} else {
			return null;  // not a literal that we can currently support
		}
		
	}

	/**
	 * Answer a String containing Java code to create a String from
	 * a Cobol Figurative Literal.
	 * <p>
	 * Supported figurative literals are:
	 * <ul>
	 * <li>SPACE, SPACES</li>
	 * <li>ZERO, ZEROS, ZEROES</li>
	 * <li>LOW-VALUE, LOW-VALUES</li>
	 * <li>HIGH-VALUE, HIGH-VALUES</li>
	 * </ul>
	 */
	private static String fromFigurativeLiteral(String valueClause, Field field) {
		int fieldLength = field.getByteLength();
		if (valueClause.startsWith("SPACE")) {
			return "StringField.makeString(" + fieldLength + ", ' ')";
		} else if (valueClause.startsWith("ZERO")) {
			if (field instanceof StringField) {
				return "StringField.makeString(" + fieldLength + ", '0')";
			} else {
				return fromNumericLiteral("0", field);
			}
		} else if (valueClause.startsWith("LOW-VALUE")) {
			return "StringField.makeString(" + fieldLength + ", '\\0')";
		} else if (valueClause.startsWith("HIGH-VALUE")) {
			return "StringField.makeString(" + fieldLength + ", StringField.HIGH_VALUE)";			
		} else {
			return null;  // an unsupported figurative literal
		}
	}
		
	/**
	 * Convert a COBOL literal String to a Java literal String
	 * @param valueClause the Cobol Literal String
	 * @return String the Java literal String, double quoted.
	 */
	private static String fromLiteralString(String valueClause) {
		if (valueClause.charAt(0) == SINGLE_QUOTE) {
			valueClause = reverseQuotes(valueClause);
		}
		StringBuffer sb = new StringBuffer();
		sb.append(DOUBLE_QUOTE);
		for (int i=1; i<valueClause.length()-1; i++) {
			char next = valueClause.charAt(i);
			if (next == DOUBLE_QUOTE) {
				sb.append('\\');
				sb.append(DOUBLE_QUOTE);
				i++; // skip the second DQ
			} else if (next == '\\') {
				sb.append('\\');
				sb.append('\\');
			} else {
				sb.append(next);
			}
		}
		sb.append(DOUBLE_QUOTE);
		return sb.toString();
	}
	

	/**
	 * Answer a String containing an Java expression that can be used
	 * to initialize a numeric field.
	 */
	private static String fromNumericLiteral(String valueClause, Field field) {
		valueClause = stripLeadingZeros(valueClause);
		if (field instanceof BigDecimalAccessor) {
			return "new BigDecimal(\"" + valueClause + "\")";
		} else if (field instanceof BigIntegerAccessor) {
			return "new BigInteger(\"" + valueClause + "\")";
		} else if (field instanceof IntAccessor
					|| field instanceof DoubleAccessor ) {   
			return valueClause; 
		} else if (field instanceof LongAccessor) {
			return valueClause + "L";
		} else if (field instanceof FloatAccessor) {
			return valueClause + "F";
		} else {
			return null;
		}
	}
	
	/**
	 * Strip any leading zeros from a numeric literal so that it 
	 * is not interpreted as an octal number.
	 */
	private static String stripLeadingZeros(String valueClause) {
		if (valueClause == null || valueClause.length() == 0) {
			return valueClause;
		}
		// must first remove any leading sign characters
		char first = valueClause.charAt(0);
		if (first == '+' || first == '-') {
			valueClause = valueClause.substring(1);
		}
		// it is not terribly efficient to use substring(), but
		// performance is not an issue here.
		while (valueClause.length() > 1 
				&& valueClause.charAt(0) == '0'
				&& Character.isDigit(valueClause.charAt(1))) {
			// only remove a leading zero if it is followed by another digit
			valueClause = valueClause.substring(1);
		}
		if (first == '-') {
			valueClause = "" + first + valueClause;
		}
		return valueClause;
	}
	
	/**
	 * Replace single quotes with double quotes and vice versa
	 */
	private static String reverseQuotes(String s) {
		StringBuffer sb = new StringBuffer();
		for (int i=0; i<s.length(); i++) {
			char next = s.charAt(i);
			if (next == SINGLE_QUOTE) {
				sb.append(DOUBLE_QUOTE);
			} else if (next == DOUBLE_QUOTE) {
				sb.append(SINGLE_QUOTE);
			} else {
				sb.append(next);
			}
		}
		return sb.toString();
	}

}
