/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

/**
 * An object that represents the OCCURS clause associated with a COBOL COPYBOOK
 * symbol.  This class represents a variable number of elements, based on another
 * field in the COPYBOOK.
 */
public class OccursDependingOn extends Occurs {

	private int min;
	private FieldSpec odoObject;

	/**
	 * Construct an instance
	 * @param min the minimum number of occurrences possible
	 * @param max the maximum number of occurrences possible
	 * @param odoObject the field that contains the (runtime) number of occurrences
	 */
	public OccursDependingOn(int min, int max, FieldSpec odoObject) {
		super(max);
		this.min = min;
		this.odoObject = odoObject;
	}
	
	/**
	 * Return the minimum number of elements that can OCCUR.
	 * @return int
	 */
	public int getMin() {
		return min;
	}

	/**
	 * Return the FieldSpec that this spec depends on to determine its number of occurrences.
	 * @return FieldSpec
	 */
	public FieldSpec getOdoObject() {
		return odoObject;
	}

	/**
	 * @see #getMin()
	 * @param min
	 */
	public void setMin(int min) {
		this.min = min;
	}

	/**
	 * @see #getOdoObject()
	 */
	public void setOdoObject(FieldSpec odoObject) {
		this.odoObject = odoObject;
	}
	
	/**
	 * A debug representation of the receiver.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(" ODO: {");
		sb.append(getMin());
		sb.append(",");
		sb.append(getMax());
		sb.append("} Object: ");
		sb.append(getOdoObject().getSymbol());
		return sb.toString();
	}

	/**
	 * Write the receiver in XML form
	 * @since 2.4.4
	 */
	@Override
	public void writeXmlAttributesOn(XMLStreamWriter writer) throws XMLStreamException {
		writer.writeAttribute("occursMin", Integer.toString(getMin()));
		writer.writeAttribute("occursMax", Integer.toString(getMax()));
		writer.writeAttribute("occursDependingOn", 
								Integer.toString(getOdoObject().getSymbolId()));
	}

	/**
	 * Parse and create/return an OccursDependingOn instance or null if none found
	 * @since 2.4.4
	 */
	public static OccursDependingOn parseXmlFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		int occursMin = parser.getAttributeIntValue("occursMin", 0);
		int occursMax = parser.getAttributeIntValue("occursMax", 0);
		int odo 	  = parser.getAttributeIntValue("occursDependingOn", 0);
		if (occursMin != 0 || occursMax != 0 || odo != 0) {
			StorageSpec ss = parser.getStorageSpecForId(odo);
			if (!(ss instanceof FieldSpec)) {
				throw new XMLStreamException("occursDependingOn=\""+odo+"\" is not a field",
						parser.getReader().getLocation());
			}
			return new OccursDependingOn(occursMin, occursMax, (FieldSpec)ss);
			
		} else {
			return null;
		}
	}

}
