/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.StringTokenizer;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import com.ibm.jzos.fields.CobolDatatypeFactory;
import com.ibm.jzos.fields.Field;
import com.ibm.jzos.recordgen.LibraryUtil;

/**
 * Generates an XML description from an Assembler CSECT or DSECT.  The input to this
 * process is an ADATA file created by the IBM High-level Assembler.
 * <p/>
 * The XML output parallels the ADATA file; for more insight as to meaning
 * of specific attributes please refer to "Enterprise COBOL for z/OS Programming Guide" /
 * "Appendix G: COBOL SYSADATA file content". 
 * <p/>
 * Attributes common to all elements:
 * 	<ul>
 * 	<li>attribute: symbol  (name)</li>
 * 	<li>attribute: level   (number)</li>
 * 	<li>attribute: symbolId (unique id; integer)</li>
 * 	<li>attribute: size    (optional; size in bytes; default=0)</li>
 * 	<li>attribute: sourceStmt (source line number)</li>
 * 	<li>attribute: offset (byte offset)</li>
 * 	<li>attribute: offsetDependingOn (optional; id of element that offset depends on)</li>
 * 	<li>attribute: symbolUnique (optional; true(default) or false)</li>
 * 	<li>attribute: varLoc  (optional; true or false(default))</li>
 * 	</ul>
 * The XML document structure consists of a top-level group element:
 * <ul>
 * <li>element: group
 * 	<ul>
 *  <li>(common attributes)</li>
 * 	<li>attribute: occurs  (optional; default 1)</li>
 * 	<li>attribute: occursMin  (optional; default 0; mutually exclusive with occurs)</li>
 * 	<li>attribute: occursMax  (optional; default 0; mutually exclusive with occurs)</li>
 * 	<li>attribute: occursDependingOn  (optional; id of element that gives occurs size of this element)</li>
 * 	<li>attribute: value      (optional; string value clause)</li>
 * 	<li>attribute: redefines  (optional; id of element redefined by this element)</li>
 * 	</ul>
 * Group elements may contain other group, field, condition, rename, or renameRange elements.
 * All elements may contain character data content consisting of the original COBOL source
 * line(s) associated with the element.
 * </li>
 * </ul>
 * Field elements are located within group elements, or in a degenerate case as a top-level solitary element.
 * <ul>
 * <li>element: field
 *  <p/>Field elements have the same attributes as group elements, with the addition of the following:
 * 	<ul>
 * 	<li>attribute: containsOdoObject  (optional; true/false(default); true if this field is the object of a OCCURS DEPENDING ON clause
 *  in another element)</li>
 *  </ul>
 *  Field elements also have attributes that direct the {@link CobolDatatypeFactory} to contruct a {@link Field} type.
 *  <p/>
 *  For String fields:
 *  <ul>
 *  <li>attribute: fieldType="String"</li>
 *  <li>attribute: length (byte length)</li>
 *  <li>attribute: trim (true/false)</li>
 *  <li>attribute: padLeft (true/false)</li>
 * 	</ul>
 *  <p/>
 *  For Binary fields:
 *  <ul>
 *  <li>attribute: fieldType="Binary"</li>
 *  <li>attribute: precision (total number of decimal digits)</li>
 *  <li>attribute: signed (true/false)</li>
 *  <li>attribute: comp5 (optional; true/false(default))</li>
 * 	</ul>
 *  <p/>
 *  For PackedDecimal fields:
 *  <ul>
 *  <li>attribute: fieldType="PackedDecimal"</li>
 *  <li>attribute: precision (total number of decimal digits)</li>
 *  <li>attribute: scale</li> *  <li>(

 *  <li>attribute: signed (true/false)</li>
 * 	</ul>
 *  <p/>
 *  For ExternalDecimal fields:
 *  <ul>
 *  <li>attribute: fieldType="ExternalDecimal"</li>
 *  <li>attribute: precision (total number of decimal digits)</li>
 *  <li>attribute: scale</li>
 *  <li>attribute: signed (true/false)</li>
 *  <li>attribute: signedTrailing (present only if signed="true"; true/false)</li>
 *  <li>attribute: signedExternal (present only if signed="true"; true/false)</li>
 *  <li>attribute: blankWhenZero  (true/false(default))</li>
 * 	</ul>
 *  <p/>
 *  For ExternalFloat fields:
 *  <ul>
 *  <li>attribute: fieldType="ExternalFloat"</li>
 *  <li>attribute: precision (total number of decimal digits)</li>
 *  <li>attribute: scale</li>
 *  <li>attribute: showMantissaPlusSign (true(default)/false)</li>
 *  <li>attribute: showExponentPlusSign (true(default)/false)</li>
 *  <li>attribute: impliedDecimal (true/false(default))</li>
 * 	</ul>
 *  <p/>
 *  For IbmFloat fields:
 *  <ul>
 *  <li>attribute: fieldType="IbmFloat"</li>
 * 	</ul>
 *  <p/>
 *  For IbmDouble fields:
 *  <ul>
 *  <li>attribute: fieldType="IbmDouble"</li>
 * 	</ul>
 *  <p/>
 *  For ByteArray fields:
 *  <ul>
 *  <li>attribute: fieldType="ByteArray"</li>
 *  <li>attribute: length (in bytes)</li>
 * 	</ul>
 * </li>
 * </ul>
 * 
 * Rename elements may be located within group or field elements:
 * <ul>
 * <li>element: rename
 * 	<ul>
 *  <li>(common attributes)</li>
 * 	<li>attribute: storage (integer id of renamed element)</li>
 * 	</ul>
 * </li>
 * </ul>
 * 
 * RenameRange elements may be located within group or field elements.
 * <ul>
 * <li>element: renameRange
 * 	<ul>
 *  <li>(common attributes)</li>
 * 	<li>attribute: fromStorage (integer id of starting element)</li>
 * 	<li>attribute: toStorage (integer id of ending element)</li>
 * 	</ul>
 * </li>
 * </ul>
 * 
 * Condition elements may be located within group or field elements:
 * <ul>
 * <li>element: condition
 * 	<ul>
 *  <li>(common attributes)</li>
 * 	<li>attribute: value (string value clause)</li>
 * 	</ul>
 * </li>
 * </ul>
 *  * <p/>
 * @see #main(String[]) main() method for information on running this class
 * @see RecordClassGenerator
 * @since 2.4.4
 */
public class RecordXMLGenerator {

	private StorageSpec specTree;
	private XMLStreamWriter writer;
	private RecordSpecTreeBuilder specTreeBuilder;
	
	/**
	 * Main method / driver for generating XML descriptions from COBOL ADATA files.
	 * <p>
	 * Arguments consist of the following optional "name=value" pairs:  
	 * <ul>
	 * <li>adatafile=//MVS.DSN  (default: //DD:SYSADATA)</li>
	 * <li>outputFile=filename  (default: writes file to System.out (stdout))</li>
	 *  <li>symbol=dataName      (default: first group found</li>
	 * </ul>
	 * <p> 
	 * If the given <code>adatafile</code> is a pathname not starting with "//", then it is assumed
	 * to be the name of a binary file in the USS (or workstation) filesystem containing the 
	 * binary SYSADATA file.  In order to process SYSADATA files on the workstation, 
	 * they must be downloaded with IBM "RDWs" included.  This can be accomplished using FTP 
	 * along with the "bin" and "quote site rdw" commands".
	 * <p> 
	 * @param args key=value strings as described above.
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		
		String adataFile = "//DD:SYSADATA";
		String outputFile = null;
		String symbol = null;
		
		for (int i = 0; i<args.length; i++) {
			StringTokenizer strtok = new StringTokenizer(args[i],"=");
			if (strtok.countTokens() != 2) {
				throw new IllegalArgumentException("Argument '" + args[i] + "' not in the form: name=value");
			}
			String name = strtok.nextToken();
			String value = strtok.nextToken();
			if ("adatafile".equalsIgnoreCase(name)) {
				adataFile = value;
			} else if ("outputfile".equalsIgnoreCase(name)) {
				outputFile = value;
			} else if ("symbol".equalsIgnoreCase(name)) {
				symbol = value;
			} else {
				throw new IllegalArgumentException("Argument '" + name + "' not recognized.");
			}
		}
		
		// Create a generator and configure it with the control arguments
		RecordXMLGenerator rxg = new RecordXMLGenerator();
		rxg.process(adataFile, symbol, outputFile);
	}
	
	/**
	 * Private ctor
	 * @see #main
	 */
	private RecordXMLGenerator() {
	}
	
	/**
	 * Construct an instance on a StorageSpec
	 */
	public RecordXMLGenerator(StorageSpec storageSpec) {
		this.specTree = storageSpec;
	}
	
	/**
	 * Process a single SYSADATA file and generate XML
	 */
	private void process(String adataFile, String symbol, String xmlFile) throws IOException {
		
		// First, a RecordSpecBuilder converts the selected symbol of the ADATA
		// file into a RecordSpec model.
		adataFile = LibraryUtil.resolveLeadingTilde(adataFile);
		specTreeBuilder = new RecordSpecTreeBuilder();
		specTreeBuilder.setSelectedSymbolName(symbol);
		specTree = specTreeBuilder.process(adataFile);
		
		generateXmlOn(xmlFile);	
	}
	
	
	public void generateXmlOn(String xmlFile) throws IOException {

		OutputStream os = null;
		try {
			if (xmlFile != null) {
				File f = new File(LibraryUtil.resolveLeadingTilde(xmlFile));
				os = new FileOutputStream(f);
			} else {
				os = System.out;
			}
			generateXmlOn(os);		
		} finally {
			try { // to close the writer
				if (os != null) os.close(); 
			} catch(Throwable ignore) {}
		}
	}
	
	public void generateXmlOn(OutputStream os) throws IOException {
		try {
			writer = makeXmlWriter(os);
			writer.writeStartDocument();
			specTree.writeXmlOn(writer, 0);
			writer.flush();
			writer.close();
		} catch (XMLStreamException xse) {
			throw new IOException(xse);
		} finally {
			os.close();
		} 
	}
	
	protected XMLStreamWriter makeXmlWriter(OutputStream os) throws XMLStreamException {
		XMLOutputFactory xof = XMLOutputFactory.newInstance();
		XMLStreamWriter writer = xof.createXMLStreamWriter(os);
		
		return writer;
	}
	
}
